/*
 * Copyright (C) 2014, Stephan Mueller <smueller@chronox.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU General Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, ALL OF
 * WHICH ARE HEREBY DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF NOT ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "cryptoperf.h"

/****************************************************************************
 * Synchronous symmetric ciphers
 ****************************************************************************/

static int cp_shash_init_test(struct cp_test *test)
{
	struct crypto_shash *shash = NULL;
	struct sdesc *desc = NULL;
	char *scratchpad = NULL;
#define MAX_KEYLEN 128
	unsigned char data[MAX_KEYLEN];
	int ret = -EFAULT;

	dbg("Initializing shash test %s\n", test->testname);
	if (!test->driver_name) {
		printk(DRIVER_NAME": missing driver_name for %s\n",
		       test->testname);
		return -EFAULT;
	}

	shash = crypto_alloc_shash(test->driver_name, 0, 0);
	if (IS_ERR(shash)) {
		printk(DRIVER_NAME": could not allocate shash handle for "
		       "%s\n", test->driver_name);
		ret = -PTR_ERR(shash);
		goto out;
	}
	desc = kmalloc(sizeof(struct sdesc) +
		       crypto_shash_descsize(shash), GFP_KERNEL);
	if (!desc) {
		ret = -ENOMEM;
		goto out;
	}
	desc->shash.flags = 0x0;
	desc->shash.tfm = shash;
	test->u.shash.desc = desc;

	/* HMAC */
	if (test->u.shash.hmac) {
		if (crypto_shash_blocksize(shash) > MAX_KEYLEN) {
			printk(DRIVER_NAME": key length for cipher %s too large %u\n",
			       test->driver_name, crypto_shash_blocksize(shash));
			goto out;
		}
		get_random_bytes(&data, crypto_shash_blocksize(shash));
		if (crypto_shash_setkey(shash, data, crypto_shash_blocksize(shash))) {
			printk(DRIVER_NAME": key could not be set\n");
			ret = -EAGAIN;
			goto out;
		}
	}


	scratchpad = kmalloc(test->u.shash.digestsize +
			     crypto_shash_blocksize(shash), GFP_KERNEL);
	if (!scratchpad) {
		printk(DRIVER_NAME": could not allocate scratchpad for "
		       "%s\n", test->driver_name);
		goto out;
	}
	get_random_bytes(scratchpad + test->u.shash.digestsize,
			 crypto_shash_blocksize(shash));

	test->u.shash.scratchpad = scratchpad;
	return 0;

out:
	if (shash)
		crypto_free_shash(shash);
	if (desc)
		kfree(desc);
	if (scratchpad)
		kfree(scratchpad);
	return ret;
}

static void cp_shash_fini_test(struct cp_test *test)
{
	dbg("Cleaning up shash test %s\n", test->testname);
	crypto_free_shash(test->u.shash.desc->shash.tfm);
	kfree(test->u.shash.scratchpad);
	kfree(test->u.shash.desc);
}

static unsigned int cp_shash_test(struct cp_test *test)
{
	crypto_shash_digest(&test->u.shash.desc->shash,
			    test->u.shash.scratchpad, test->u.shash.digestsize,
			    test->u.shash.scratchpad + test->u.shash.digestsize);
	return crypto_shash_blocksize(test->u.shash.desc->shash.tfm);
}

struct cp_shash_tests {
	char *testname;
	char *driver_name;
	unsigned int digestsize;
	unsigned int hmac;
};

static const struct cp_shash_tests testcases[] = {
	{ "SHA-1(G)", "sha1-generic", 20, 0 },
	{ "SHA-224(G)", "sha224-generic", 28, 0 },
	{ "SHA-256(G)", "sha256-generic", 32, 0 },
	{ "SHA-384(G)", "sha384-generic", 48, 0 },
	{ "SHA-512(G)", "sha512-generic", 64, 0 },
	{ "SHA-1(SSSE3 AVX AVX2)", "sha1-ssse3", 20, 0 },
	{ "SHA-224(SSSE3 AVX AVX2)", "sha224-ssse3", 28, 0 },
	{ "SHA-256(SSSE3 AVX AVX2)", "sha256-ssse3", 32, 0 },
	{ "SHA-384(SSSE3 AVX AVX2)", "sha384-ssse3", 48, 0 },
	{ "SHA-512(SSSE3 AVX AVX2)", "sha512-ssse3", 64, 0 },

	{ "HMAC SHA-1(G)", "hmac(sha1-generic)", 20, 1 },
	{ "HMAC SHA-224(G)", "hmac(sha224-generic)", 28, 1 },
	{ "HMAC SHA-256(G)", "hmac(sha256-generic)", 32, 1 },
	{ "HMAC SHA-384(G)", "hmac(sha384-generic)", 48, 1 },
	{ "HMAC SHA-512(G)", "hmac(sha512-generic)", 64, 1 },
	{ "HMAC SHA-1(SSSE3 AVX AVX2)", "hmac(sha1-ssse3)", 20, 1 },
	{ "HMAC SHA-224(SSSE3 AVX AVX2)", "hmac(sha224-ssse3)", 28, 1 },
	{ "HMAC SHA-256(SSSE3 AVX AVX2)", "hmac(sha256-ssse3)", 32, 1 },
	{ "HMAC SHA-384(SSSE3 AVX AVX2)", "hmac(sha384-ssse3)", 48, 1 },
	{ "HMAC SHA-512(SSSE3 AVX AVX2)", "hmac(sha512-ssse3)", 64, 1 },

	{ "MD5(G)", "md5-generic", 16, 0 },
	{ "HMAC MD5(G)", "hmac(md5-generic)", 16, 1 },
};

static struct cp_test cp_shash_testdef[(ARRAY_SIZE(testcases))];

int cp_shash_register(void)
{
	size_t i = 0;

	for (i = i; i < ARRAY_SIZE(testcases); i++) {
		int ret = 0;
		cp_shash_testdef[i].testname = testcases[i].testname;
		cp_shash_testdef[i].driver_name = testcases[i].driver_name;
		cp_shash_testdef[i].type = "shash";
		cp_shash_testdef[i].exectime = DFLT_EXECTIME;
		cp_shash_testdef[i].u.shash.hmac = testcases[i].hmac;
		cp_shash_testdef[i].u.shash.digestsize = testcases[i].digestsize;
		cp_shash_testdef[i].init_test = cp_shash_init_test;
		cp_shash_testdef[i].fini_test = cp_shash_fini_test;
		cp_shash_testdef[i].exec_test = cp_shash_test;
		ret = cp_register_test(&cp_shash_testdef[i]);
		if (ret)
			return ret;
	}
	return 0;
}
